<?php

/*
 * PHP Utilities - Misc GS1 functions
 * Copyright 2023 Daniel Marschall, ViaThinkSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Shows the GS1 Application Identifier for a given GS1 barcode
 * @param string $gs1 The GS1 barcode, e.g. 0012000000171
 * @return string The detected GS1 Application Identifier
 */
function gs1_barcode_show_appidentifier(string $gs1): string {
	static $aids = null;
	if ($aids === null) {
		// Lookup table taken from
		// https://www.gs1.org/standards/barcodes/application-identifiers?lang=en
		// Date: 26 March 2023 (519 entries)

		// The following code is used to re-generate the lookup table $aids ($raw is inserted via copy-paste from the browser)
		/*
		$raw = <<<EOF
		AI	Description	Format	Data title	FNC1 required?	Regular expression
		00	Serial Shipping Container Code (SSCC)	N2+N18	SSCC	No	^00(\d{18})$
		01	Global Trade Item Number (GTIN)	N2+N14	GTIN	No	^01(\d{14})$
		.......
		EOF;
		foreach (explode("\n",$raw) as $line) {
			$line = trim($line);
			if ($line == '') continue;
			if (strpos($line,"\t") === false) continue;
			list($prefix, $description) = explode("\t", $line);
			if ($prefix == "AI") continue;
			$description = str_replace('\\', '\\\\', $description);
			$description = str_replace("'", "\\'", $description);
			echo "\$aids['$prefix'] = '$description';\n";
		}
		*/

		$aids['00'] = 'Serial Shipping Container Code (SSCC)';
		$aids['01'] = 'Global Trade Item Number (GTIN)';
		$aids['02'] = 'Global Trade Item Number (GTIN) of contained trade items';
		$aids['10'] = 'Batch or lot number';
		$aids['11'] = 'Production date (YYMMDD)';
		$aids['12'] = 'Due date (YYMMDD)';
		$aids['13'] = 'Packaging date (YYMMDD)';
		$aids['15'] = 'Best before date (YYMMDD)';
		$aids['16'] = 'Sell by date (YYMMDD)';
		$aids['17'] = 'Expiration date (YYMMDD)';
		$aids['20'] = 'Internal product variant';
		$aids['21'] = 'Serial number';
		$aids['22'] = 'Consumer product variant';
		$aids['235'] = 'Third Party Controlled, Serialised Extension of Global Trade Item Number (GTIN) (TPX)';
		$aids['240'] = 'Additional product identification assigned by the manufacturer';
		$aids['241'] = 'Customer part number';
		$aids['242'] = 'Made-to-Order variation number';
		$aids['243'] = 'Packaging component number';
		$aids['250'] = 'Secondary serial number';
		$aids['251'] = 'Reference to source entity';
		$aids['253'] = 'Global Document Type Identifier (GDTI)';
		$aids['254'] = 'Global Location Number (GLN) extension component';
		$aids['255'] = 'Global Coupon Number (GCN)';
		$aids['30'] = 'Variable count of items (variable measure trade item)';
		$aids['3100'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3101'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3102'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3103'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3104'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3105'] = 'Net weight, kilograms (variable measure trade item)';
		$aids['3110'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3111'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3112'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3113'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3114'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3115'] = 'Length or first dimension, metres (variable measure trade item)';
		$aids['3120'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3121'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3122'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3123'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3124'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3125'] = 'Width, diameter, or second dimension, metres (variable measure trade item)';
		$aids['3130'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3131'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3132'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3133'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3134'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3135'] = 'Depth, thickness, height, or third dimension, metres (variable measure trade item)';
		$aids['3140'] = 'Area, square metres (variable measure trade item)';
		$aids['3141'] = 'Area, square metres (variable measure trade item)';
		$aids['3142'] = 'Area, square metres (variable measure trade item)';
		$aids['3143'] = 'Area, square metres (variable measure trade item)';
		$aids['3144'] = 'Area, square metres (variable measure trade item)';
		$aids['3145'] = 'Area, square metres (variable measure trade item)';
		$aids['3150'] = 'Net volume, litres (variable measure trade item)';
		$aids['3151'] = 'Net volume, litres (variable measure trade item)';
		$aids['3152'] = 'Net volume, litres (variable measure trade item)';
		$aids['3153'] = 'Net volume, litres (variable measure trade item)';
		$aids['3154'] = 'Net volume, litres (variable measure trade item)';
		$aids['3155'] = 'Net volume, litres (variable measure trade item)';
		$aids['3160'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3161'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3162'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3163'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3164'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3165'] = 'Net volume, cubic metres (variable measure trade item)';
		$aids['3200'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3201'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3202'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3203'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3204'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3205'] = 'Net weight, pounds (variable measure trade item)';
		$aids['3210'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3211'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3212'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3213'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3214'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3215'] = 'Length or first dimension, inches (variable measure trade item)';
		$aids['3220'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3221'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3222'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3223'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3224'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3225'] = 'Length or first dimension, feet (variable measure trade item)';
		$aids['3230'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3231'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3232'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3233'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3234'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3235'] = 'Length or first dimension, yards (variable measure trade item)';
		$aids['3240'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3241'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3242'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3243'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3244'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3245'] = 'Width, diameter, or second dimension, inches (variable measure trade item)';
		$aids['3250'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3251'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3252'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3253'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3254'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3255'] = 'Width, diameter, or second dimension, feet (variable measure trade item)';
		$aids['3260'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3261'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3262'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3263'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3264'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3265'] = 'Width, diameter, or second dimension, yards (variable measure trade item)';
		$aids['3270'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3271'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3272'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3273'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3274'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3275'] = 'Depth, thickness, height, or third dimension, inches (variable measure trade item)';
		$aids['3280'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3281'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3282'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3283'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3284'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3285'] = 'Depth, thickness, height, or third dimension, feet (variable measure trade item)';
		$aids['3290'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3291'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3292'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3293'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3294'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3295'] = 'Depth, thickness, height, or third dimension, yards (variable measure trade item)';
		$aids['3300'] = 'Logistic weight, kilograms';
		$aids['3301'] = 'Logistic weight, kilograms';
		$aids['3302'] = 'Logistic weight, kilograms';
		$aids['3303'] = 'Logistic weight, kilograms';
		$aids['3304'] = 'Logistic weight, kilograms';
		$aids['3305'] = 'Logistic weight, kilograms';
		$aids['3310'] = 'Length or first dimension, metres';
		$aids['3311'] = 'Length or first dimension, metres';
		$aids['3312'] = 'Length or first dimension, metres';
		$aids['3313'] = 'Length or first dimension, metres';
		$aids['3314'] = 'Length or first dimension, metres';
		$aids['3315'] = 'Length or first dimension, metres';
		$aids['3320'] = 'Width, diameter, or second dimension, metres';
		$aids['3321'] = 'Width, diameter, or second dimension, metres';
		$aids['3322'] = 'Width, diameter, or second dimension, metres';
		$aids['3323'] = 'Width, diameter, or second dimension, metres';
		$aids['3324'] = 'Width, diameter, or second dimension, metres';
		$aids['3325'] = 'Width, diameter, or second dimension, metres';
		$aids['3330'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3331'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3332'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3333'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3334'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3335'] = 'Depth, thickness, height, or third dimension, metres';
		$aids['3340'] = 'Area, square metres';
		$aids['3341'] = 'Area, square metres';
		$aids['3342'] = 'Area, square metres';
		$aids['3343'] = 'Area, square metres';
		$aids['3344'] = 'Area, square metres';
		$aids['3345'] = 'Area, square metres';
		$aids['3350'] = 'Logistic volume, litres';
		$aids['3351'] = 'Logistic volume, litres';
		$aids['3352'] = 'Logistic volume, litres';
		$aids['3353'] = 'Logistic volume, litres';
		$aids['3354'] = 'Logistic volume, litres';
		$aids['3355'] = 'Logistic volume, litres';
		$aids['3360'] = 'Logistic volume, cubic metres';
		$aids['3361'] = 'Logistic volume, cubic metres';
		$aids['3362'] = 'Logistic volume, cubic metres';
		$aids['3363'] = 'Logistic volume, cubic metres';
		$aids['3364'] = 'Logistic volume, cubic metres';
		$aids['3365'] = 'Logistic volume, cubic metres';
		$aids['3370'] = 'Kilograms per square metre';
		$aids['3371'] = 'Kilograms per square metre';
		$aids['3372'] = 'Kilograms per square metre';
		$aids['3373'] = 'Kilograms per square metre';
		$aids['3374'] = 'Kilograms per square metre';
		$aids['3375'] = 'Kilograms per square metre';
		$aids['3400'] = 'Logistic weight, pounds';
		$aids['3401'] = 'Logistic weight, pounds';
		$aids['3402'] = 'Logistic weight, pounds';
		$aids['3403'] = 'Logistic weight, pounds';
		$aids['3404'] = 'Logistic weight, pounds';
		$aids['3405'] = 'Logistic weight, pounds';
		$aids['3410'] = 'Length or first dimension, inches';
		$aids['3411'] = 'Length or first dimension, inches';
		$aids['3412'] = 'Length or first dimension, inches';
		$aids['3413'] = 'Length or first dimension, inches';
		$aids['3414'] = 'Length or first dimension, inches';
		$aids['3415'] = 'Length or first dimension, inches';
		$aids['3420'] = 'Length or first dimension, feet';
		$aids['3421'] = 'Length or first dimension, feet';
		$aids['3422'] = 'Length or first dimension, feet';
		$aids['3423'] = 'Length or first dimension, feet';
		$aids['3424'] = 'Length or first dimension, feet';
		$aids['3425'] = 'Length or first dimension, feet';
		$aids['3430'] = 'Length or first dimension, yards';
		$aids['3431'] = 'Length or first dimension, yards';
		$aids['3432'] = 'Length or first dimension, yards';
		$aids['3433'] = 'Length or first dimension, yards';
		$aids['3434'] = 'Length or first dimension, yards';
		$aids['3435'] = 'Length or first dimension, yards';
		$aids['3440'] = 'Width, diameter, or second dimension, inches';
		$aids['3441'] = 'Width, diameter, or second dimension, inches';
		$aids['3442'] = 'Width, diameter, or second dimension, inches';
		$aids['3443'] = 'Width, diameter, or second dimension, inches';
		$aids['3444'] = 'Width, diameter, or second dimension, inches';
		$aids['3445'] = 'Width, diameter, or second dimension, inches';
		$aids['3450'] = 'Width, diameter, or second dimension, feet';
		$aids['3451'] = 'Width, diameter, or second dimension, feet';
		$aids['3452'] = 'Width, diameter, or second dimension, feet';
		$aids['3453'] = 'Width, diameter, or second dimension, feet';
		$aids['3454'] = 'Width, diameter, or second dimension, feet';
		$aids['3455'] = 'Width, diameter, or second dimension, feet';
		$aids['3460'] = 'Width, diameter, or second dimension, yard';
		$aids['3461'] = 'Width, diameter, or second dimension, yard';
		$aids['3462'] = 'Width, diameter, or second dimension, yard';
		$aids['3463'] = 'Width, diameter, or second dimension, yard';
		$aids['3464'] = 'Width, diameter, or second dimension, yard';
		$aids['3465'] = 'Width, diameter, or second dimension, yard';
		$aids['3470'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3471'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3472'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3473'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3474'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3475'] = 'Depth, thickness, height, or third dimension, inches';
		$aids['3480'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3481'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3482'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3483'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3484'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3485'] = 'Depth, thickness, height, or third dimension, feet';
		$aids['3490'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3491'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3492'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3493'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3494'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3495'] = 'Depth, thickness, height, or third dimension, yards';
		$aids['3500'] = 'Area, square inches (variable measure trade item)';
		$aids['3501'] = 'Area, square inches (variable measure trade item)';
		$aids['3502'] = 'Area, square inches (variable measure trade item)';
		$aids['3503'] = 'Area, square inches (variable measure trade item)';
		$aids['3504'] = 'Area, square inches (variable measure trade item)';
		$aids['3505'] = 'Area, square inches (variable measure trade item)';
		$aids['3510'] = 'Area, square feet (variable measure trade item)';
		$aids['3511'] = 'Area, square feet (variable measure trade item)';
		$aids['3512'] = 'Area, square feet (variable measure trade item)';
		$aids['3513'] = 'Area, square feet (variable measure trade item)';
		$aids['3514'] = 'Area, square feet (variable measure trade item)';
		$aids['3515'] = 'Area, square feet (variable measure trade item)';
		$aids['3520'] = 'Area, square yards (variable measure trade item)';
		$aids['3521'] = 'Area, square yards (variable measure trade item)';
		$aids['3522'] = 'Area, square yards (variable measure trade item)';
		$aids['3523'] = 'Area, square yards (variable measure trade item)';
		$aids['3524'] = 'Area, square yards (variable measure trade item)';
		$aids['3525'] = 'Area, square yards (variable measure trade item)';
		$aids['3530'] = 'Area, square inches';
		$aids['3531'] = 'Area, square inches';
		$aids['3532'] = 'Area, square inches';
		$aids['3533'] = 'Area, square inches';
		$aids['3534'] = 'Area, square inches';
		$aids['3535'] = 'Area, square inches';
		$aids['3540'] = 'Area, square feet';
		$aids['3541'] = 'Area, square feet';
		$aids['3542'] = 'Area, square feet';
		$aids['3543'] = 'Area, square feet';
		$aids['3544'] = 'Area, square feet';
		$aids['3545'] = 'Area, square feet';
		$aids['3550'] = 'Area, square yards';
		$aids['3551'] = 'Area, square yards';
		$aids['3552'] = 'Area, square yards';
		$aids['3553'] = 'Area, square yards';
		$aids['3554'] = 'Area, square yards';
		$aids['3555'] = 'Area, square yards';
		$aids['3560'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3561'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3562'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3563'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3564'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3565'] = 'Net weight, troy ounces (variable measure trade item)';
		$aids['3570'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3571'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3572'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3573'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3574'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3575'] = 'Net weight (or volume), ounces (variable measure trade item)';
		$aids['3600'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3601'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3602'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3603'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3604'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3605'] = 'Net volume, quarts (variable measure trade item)';
		$aids['3610'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3611'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3612'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3613'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3614'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3615'] = 'Net volume, gallons U.S. (variable measure trade item)';
		$aids['3620'] = 'Logistic volume, quarts';
		$aids['3621'] = 'Logistic volume, quarts';
		$aids['3622'] = 'Logistic volume, quarts';
		$aids['3623'] = 'Logistic volume, quarts';
		$aids['3624'] = 'Logistic volume, quarts';
		$aids['3625'] = 'Logistic volume, quarts';
		$aids['3630'] = 'Logistic volume, gallons U.S.';
		$aids['3631'] = 'Logistic volume, gallons U.S.';
		$aids['3632'] = 'Logistic volume, gallons U.S.';
		$aids['3633'] = 'Logistic volume, gallons U.S.';
		$aids['3634'] = 'Logistic volume, gallons U.S.';
		$aids['3635'] = 'Logistic volume, gallons U.S.';
		$aids['3640'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3641'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3642'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3643'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3644'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3645'] = 'Net volume, cubic inches (variable measure trade item)';
		$aids['3650'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3651'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3652'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3653'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3654'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3655'] = 'Net volume, cubic feet (variable measure trade item)';
		$aids['3660'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3661'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3662'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3663'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3664'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3665'] = 'Net volume, cubic yards (variable measure trade item)';
		$aids['3670'] = 'Logistic volume, cubic inches';
		$aids['3671'] = 'Logistic volume, cubic inches';
		$aids['3672'] = 'Logistic volume, cubic inches';
		$aids['3673'] = 'Logistic volume, cubic inches';
		$aids['3674'] = 'Logistic volume, cubic inches';
		$aids['3675'] = 'Logistic volume, cubic inches';
		$aids['3680'] = 'Logistic volume, cubic feet';
		$aids['3681'] = 'Logistic volume, cubic feet';
		$aids['3682'] = 'Logistic volume, cubic feet';
		$aids['3683'] = 'Logistic volume, cubic feet';
		$aids['3684'] = 'Logistic volume, cubic feet';
		$aids['3685'] = 'Logistic volume, cubic feet';
		$aids['3690'] = 'Logistic volume, cubic yards';
		$aids['3691'] = 'Logistic volume, cubic yards';
		$aids['3692'] = 'Logistic volume, cubic yards';
		$aids['3693'] = 'Logistic volume, cubic yards';
		$aids['3694'] = 'Logistic volume, cubic yards';
		$aids['3695'] = 'Logistic volume, cubic yards';
		$aids['37'] = 'Count of trade items or trade item pieces contained in a logistic unit';
		$aids['3900'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3901'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3902'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3903'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3904'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3905'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3906'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3907'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3908'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3909'] = 'Applicable amount payable or Coupon value, local currency';
		$aids['3910'] = 'Applicable amount payable with ISO currency code';
		$aids['3911'] = 'Applicable amount payable with ISO currency code';
		$aids['3912'] = 'Applicable amount payable with ISO currency code';
		$aids['3913'] = 'Applicable amount payable with ISO currency code';
		$aids['3914'] = 'Applicable amount payable with ISO currency code';
		$aids['3915'] = 'Applicable amount payable with ISO currency code';
		$aids['3916'] = 'Applicable amount payable with ISO currency code';
		$aids['3917'] = 'Applicable amount payable with ISO currency code';
		$aids['3918'] = 'Applicable amount payable with ISO currency code';
		$aids['3919'] = 'Applicable amount payable with ISO currency code';
		$aids['3920'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3921'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3922'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3923'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3924'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3925'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3926'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3927'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3928'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3929'] = 'Applicable amount payable, single monetary area (variable measure trade item)';
		$aids['3930'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3931'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3932'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3933'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3934'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3935'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3936'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3937'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3938'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3939'] = 'Applicable amount payable with ISO currency code (variable measure trade item)';
		$aids['3940'] = 'Percentage discount of a coupon';
		$aids['3941'] = 'Percentage discount of a coupon';
		$aids['3942'] = 'Percentage discount of a coupon';
		$aids['3943'] = 'Percentage discount of a coupon';
		$aids['3950'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3951'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3952'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3953'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3954'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3955'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3956'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3957'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3958'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['3959'] = 'Amount Payable per unit of measure single monetary area (variable measure trade item)';
		$aids['400'] = 'Customers purchase order number';
		$aids['401'] = 'Global Identification Number for Consignment (GINC)';
		$aids['402'] = 'Global Shipment Identification Number (GSIN)';
		$aids['403'] = 'Routing code';
		$aids['410'] = 'Ship to / Deliver to Global Location Number (GLN)';
		$aids['411'] = 'Bill to / Invoice to Global Location Number (GLN)';
		$aids['412'] = 'Purchased from Global Location Number (GLN)';
		$aids['413'] = 'Ship for / Deliver for - Forward to Global Location Number (GLN)';
		$aids['414'] = 'Identification of a physical location - Global Location Number (GLN)';
		$aids['415'] = 'Global Location Number (GLN) of the invoicing party';
		$aids['416'] = 'Global Location Number (GLN) of the production or service location';
		$aids['417'] = 'Party Global Location Number (GLN)';
		$aids['420'] = 'Ship to / Deliver to postal code within a single postal authority';
		$aids['421'] = 'Ship to / Deliver to postal code with ISO country code';
		$aids['422'] = 'Country of origin of a trade item';
		$aids['423'] = 'Country of initial processing';
		$aids['424'] = 'Country of processing';
		$aids['425'] = 'Country of disassembly';
		$aids['426'] = 'Country covering full process chain';
		$aids['427'] = 'Country subdivision Of origin';
		$aids['4300'] = 'Ship-to / Deliver-to company name';
		$aids['4301'] = 'Ship-to / Deliver-to contact';
		$aids['4302'] = 'Ship-to / Deliver-to address line 1';
		$aids['4303'] = 'Ship-to / Deliver-to address line 2';
		$aids['4304'] = 'Ship-to / Deliver-to suburb';
		$aids['4305'] = 'Ship-to / Deliver-to locality';
		$aids['4306'] = 'Ship-to / Deliver-to region';
		$aids['4307'] = 'Ship-to / Deliver-to country code';
		$aids['4308'] = 'Ship-to / Deliver-to telephone number';
		$aids['4309'] = 'Ship-to / Deliver-to GEO location';
		$aids['4310'] = 'Return-to company name';
		$aids['4311'] = 'Return-to contact';
		$aids['4312'] = 'Return-to address line 1';
		$aids['4313'] = 'Return-to address line 2';
		$aids['4314'] = 'Return-to suburb';
		$aids['4315'] = 'Return-to locality';
		$aids['4316'] = 'Return-to region';
		$aids['4317'] = 'Return-to country code';
		$aids['4318'] = 'Return-to postal code';
		$aids['4319'] = 'Return-to telephone number';
		$aids['4320'] = 'Service code description';
		$aids['4321'] = 'Dangerous goods flag';
		$aids['4322'] = 'Authority to leave';
		$aids['4323'] = 'Signature required flag';
		$aids['4324'] = 'Not before delivery date time';
		$aids['4325'] = 'Not after delivery date time';
		$aids['4326'] = 'Release date';
		$aids['7001'] = 'NATO Stock Number (NSN)';
		$aids['7002'] = 'UN/ECE meat carcasses and cuts classification';
		$aids['7003'] = 'Expiration date and time';
		$aids['7004'] = 'Active potency';
		$aids['7005'] = 'Catch area';
		$aids['7006'] = 'First freeze date';
		$aids['7007'] = 'Harvest date';
		$aids['7008'] = 'Species for fishery purposes';
		$aids['7009'] = 'Fishing gear type';
		$aids['7010'] = 'Production method';
		$aids['7011'] = 'Test by date';
		$aids['7020'] = 'Refurbishment lot ID';
		$aids['7021'] = 'Functional status';
		$aids['7022'] = 'Revision status';
		$aids['7023'] = 'Global Individual Asset Identifier (GIAI) of an assembly';
		$aids['7030'] = 'Number of processor with ISO Country Code';
		$aids['7031'] = 'Number of processor with ISO Country Code';
		$aids['7032'] = 'Number of processor with ISO Country Code';
		$aids['7033'] = 'Number of processor with ISO Country Code';
		$aids['7034'] = 'Number of processor with ISO Country Code';
		$aids['7035'] = 'Number of processor with ISO Country Code';
		$aids['7036'] = 'Number of processor with ISO Country Code';
		$aids['7037'] = 'Number of processor with ISO Country Code';
		$aids['7038'] = 'Number of processor with ISO Country Code';
		$aids['7039'] = 'Number of processor with ISO Country Code';
		$aids['7040'] = 'GS1 UIC with Extension 1 and Importer index';
		$aids['710'] = 'National Healthcare Reimbursement Number (NHRN) - Germany PZN';
		$aids['711'] = 'National Healthcare Reimbursement Number (NHRN) - France CIP';
		$aids['712'] = 'National Healthcare Reimbursement Number (NHRN) - Spain CN';
		$aids['713'] = 'National Healthcare Reimbursement Number (NHRN) - Brasil DRN';
		$aids['714'] = 'National Healthcare Reimbursement Number (NHRN) - Portugal AIM';
		$aids['715'] = 'National Healthcare Reimbursement Number (NHRN) - United States of America NDC';
		$aids['7230'] = 'Certification reference';
		$aids['7231'] = 'Certification reference';
		$aids['7232'] = 'Certification reference';
		$aids['7233'] = 'Certification reference';
		$aids['7234'] = 'Certification reference';
		$aids['7235'] = 'Certification reference';
		$aids['7236'] = 'Certification reference';
		$aids['7237'] = 'Certification reference';
		$aids['7238'] = 'Certification reference';
		$aids['7239'] = 'Certification reference';
		$aids['7240'] = 'Protocol ID';
		$aids['8001'] = 'Roll products (width, length, core diameter, direction, splices)';
		$aids['8002'] = 'Cellular mobile telephone identifier';
		$aids['8003'] = 'Global Returnable Asset Identifier (GRAI)';
		$aids['8004'] = 'Global Individual Asset Identifier (GIAI)';
		$aids['8005'] = 'Price per unit of measure';
		$aids['8006'] = 'Identification of an individual trade item piece (ITIP)';
		$aids['8007'] = 'International Bank Account Number (IBAN)';
		$aids['8008'] = 'Date and time of production';
		$aids['8009'] = 'Optically Readable Sensor Indicator';
		$aids['8010'] = 'Component/Part Identifier (CPID)';
		$aids['8011'] = 'Component/Part Identifier serial number (CPID SERIAL)';
		$aids['8012'] = 'Software version';
		$aids['8013'] = 'Global Model Number (GMN)';
		$aids['8017'] = 'Global Service Relation Number (GSRN) to identify the relationship between an organisation offering services and the provider of services';
		$aids['8018'] = 'Global Service Relation Number (GSRN) to identify the relationship between an organisation offering services and the recipient of services';
		$aids['8019'] = 'Service Relation Instance Number (SRIN)';
		$aids['8020'] = 'Payment slip reference number';
		$aids['8026'] = 'Identification of pieces of a trade item (ITIP) contained in a logistic unit';
		$aids['8110'] = 'Coupon code identification for use in North America';
		$aids['8111'] = 'Loyalty points of a coupon';
		$aids['8112'] = 'Paperless coupon code identification for use in North America';
		$aids['8200'] = 'Extended Packaging URL';
		$aids['90'] = 'Information mutually agreed between trading partners';
		$aids['91'] = 'Company internal information';
		$aids['92'] = 'Company internal information';
		$aids['93'] = 'Company internal information';
		$aids['94'] = 'Company internal information';
		$aids['95'] = 'Company internal information';
		$aids['96'] = 'Company internal information';
		$aids['97'] = 'Company internal information';
		$aids['98'] = 'Company internal information';
		$aids['99'] = 'Company internal information';
	}

	foreach ($aids as $aid => $description) {
		if (strpos($gs1, $aid) === 0) {
			return "$aid = $description";
		}
	}
	return "Unknown";
}